import uuid
from dataclasses import replace

from taoverse.model.competition.data import ModelConstraints
from taoverse.model.data import Model, ModelId
from taoverse.model.storage.disk import utils
from taoverse.model.storage.remote_model_store import RemoteModelStore


class FakeRemoteModelStore(RemoteModelStore):
    """Fake implementation for remotely storing and retrieving a model."""

    def __init__(self):
        self.remote_models = dict()

    async def upload_model(
        self, model: Model, model_constraints: ModelConstraints
    ) -> ModelId:
        """Fake uploads a model."""

        model_id = model.id
        # Generate a commit and hash, if one doesn't yet exist.
        if model_id.commit is None:
            commit = str(uuid.uuid4())
            model_id = replace(model_id, commit=commit)

        if model_id.hash is None:
            hash = str(uuid.uuid4())
            model_id = replace(model_id, hash=hash)

        model = replace(model, id=model_id)

        self.remote_models[model_id] = model

        return model_id

    async def download_model(
        self, model_id: ModelId, local_path: str, model_constraints: ModelConstraints
    ) -> Model:
        """Retrieves a trained model from memory."""

        model = self.remote_models[model_id]

        # Parse out the hotkey and the base path from local_path to replicate hugging face logic.
        split_string = local_path.split("/")

        # Store it at the local_path
        dir = utils.get_local_model_snapshot_dir(
            split_string[0], split_string[2], model_id
        )
        model.pt_model.save_pretrained(
            save_directory=dir,
            safe_serialization=True,
        )

        return model

    def inject_mismatched_model(self, model_id: ModelId, model: Model) -> ModelId:
        """Fake uploads a model by a specific model id."""

        # Use provided commit + hash rather than generating a new one.
        self.remote_models[model_id] = model

        return model_id

    def get_only_model(self) -> Model:
        """Returns the only uploaded model or raises a ValueError if none or more than one is found."""

        if len(self.remote_models) != 1:
            raise ValueError(
                f"There are {len(self.remote_models)} uploaded models. Expected 1."
            )

        return [i for i in self.remote_models.values()][0]

    def reset(self):
        """Resets the store to its initial state."""
        self.remote_models = dict()



import math
import os
import random
import unittest
from dataclasses import replace

from taoverse.model.data import EvalResult, ModelId, ModelMetadata
from taoverse.model.model_tracker import ModelTracker


class TestModelTracker(unittest.TestCase):
    def setUp(self):
        self.model_tracker = ModelTracker()

    def _create_model_id(self) -> ModelId:
        return ModelId(
            namespace="test_model",
            name="test_name",
            commit="test_commit",
            hash="test_hash",
            secure_hash="test_secure_hash",
            competition_id=1,
        )

    def test_roundtrip_state(self):
        hotkey = "test_hotkey"
        model_id = self._create_model_id()
        model_metadata = ModelMetadata(id=model_id, block=1)

        state_path = ".test_tracker_state.pickle"
        self.model_tracker.on_miner_model_updated(hotkey, model_metadata)
        self.model_tracker.on_model_evaluated(
            hotkey,
            1,
            EvalResult(
                block=2, score=math.inf, winning_model_block=1, winning_model_score=2
            ),
        )
        self.model_tracker.on_model_evaluated(
            hotkey,
            1,
            EvalResult(
                block=3, score=0.5, winning_model_block=1, winning_model_score=2
            ),
        )
        self.model_tracker.save_state(state_path)

        new_tracker = ModelTracker()
        new_tracker.load_state(state_path)

        os.remove(state_path)

        self.assertEqual(
            self.model_tracker.miner_hotkey_to_model_metadata_dict,
            new_tracker.miner_hotkey_to_model_metadata_dict,
        )
        self.assertEqual(
            self.model_tracker.miner_hotkey_to_eval_results,
            new_tracker.miner_hotkey_to_eval_results,
        )

    def test_on_miner_model_updated_add(self):
        hotkey = "test_hotkey"
        model_id = self._create_model_id()

        model_metadata = ModelMetadata(id=model_id, block=1)

        self.model_tracker.on_miner_model_updated(hotkey, model_metadata)

        self.assertTrue(
            hotkey in self.model_tracker.miner_hotkey_to_model_metadata_dict
        )
        self.assertEqual(
            model_metadata,
            self.model_tracker.miner_hotkey_to_model_metadata_dict[hotkey],
        )

    def test_on_miner_model_updated_update(self):
        hotkey = "test_hotkey"
        model_id = self._create_model_id()

        model_metadata = ModelMetadata(id=model_id, block=1)

        new_model_id = replace(
            model_id,
            commit="other_commit",
            hash="other_hash",
            secure_hash="other_secure_hash",
        )
        new_model_metadata = ModelMetadata(id=new_model_id, block=2)

        self.model_tracker.on_miner_model_updated(hotkey, model_metadata)
        self.model_tracker.on_miner_model_updated(hotkey, new_model_metadata)

        self.assertTrue(
            hotkey in self.model_tracker.miner_hotkey_to_model_metadata_dict
        )
        self.assertEqual(
            new_model_metadata,
            self.model_tracker.miner_hotkey_to_model_metadata_dict[hotkey],
        )

    def test_get_model_metadata_for_miner_hotkey(self):
        hotkey = "test_hotkey"
        model_id = model_id = self._create_model_id()

        model_metadata = ModelMetadata(id=model_id, block=1)

        self.model_tracker.on_miner_model_updated(hotkey, model_metadata)
        returned_model_metadata = (
            self.model_tracker.get_model_metadata_for_miner_hotkey(hotkey)
        )

        self.assertEqual(model_metadata, returned_model_metadata)

    def test_get_model_metadata_for_miner_hotkey_optional(self):
        hotkey = "test_hotkey"

        returned_model_id = self.model_tracker.get_model_metadata_for_miner_hotkey(
            hotkey
        )

        self.assertIsNone(returned_model_id)

    def test_get_miner_hotkey_to_model_metadata_dict(self):
        hotkey_1 = "test_hotkey"
        model_id_1 = ModelId(
            namespace="test_model",
            name="test_name",
            commit="test_commit",
            hash="test_hash",
            secure_hash="test_secure_hash",
            competition_id=1,
        )
        model_metadata_1 = ModelMetadata(id=model_id_1, block=1)

        hotkey_2 = "test_hotkey2"
        model_id_2 = ModelId(
            namespace="test_model2",
            name="test_name2",
            commit="test_commit2",
            hash="test_hash2",
            secure_hash="test_secure_hash2",
            competition_id=1,
        )
        model_metadata_2 = ModelMetadata(id=model_id_2, block=2)

        self.model_tracker.on_miner_model_updated(hotkey_1, model_metadata_1)
        self.model_tracker.on_miner_model_updated(hotkey_2, model_metadata_2)

        hotkey_to_model_metadata = (
            self.model_tracker.get_miner_hotkey_to_model_metadata_dict()
        )

        self.assertEqual(len(hotkey_to_model_metadata), 2)
        self.assertEqual(hotkey_to_model_metadata[hotkey_1], model_metadata_1)
        self.assertEqual(hotkey_to_model_metadata[hotkey_2], model_metadata_2)

    def test_on_hotkeys_updated_extra_ignored(self):
        hotkey = "test_hotkey"
        model_id = self._create_model_id()

        model_metadata = ModelMetadata(id=model_id, block=1)

        self.model_tracker.on_miner_model_updated(hotkey, model_metadata)
        self.model_tracker.on_hotkeys_updated(set([hotkey, "extra_hotkey"]))

        self.assertEqual(len(self.model_tracker.miner_hotkey_to_model_metadata_dict), 1)

    def test_on_hotkeys_updated_missing_removed(self):
        hotkey = "test_hotkey"
        model_id = self._create_model_id()

        model_metadata = ModelMetadata(id=model_id, block=1)

        self.model_tracker.on_miner_model_updated(hotkey, model_metadata)
        self.model_tracker.on_model_evaluated(
            hotkey,
            1,
            EvalResult(block=2, score=1, winning_model_block=1, winning_model_score=2),
        )
        self.model_tracker.on_hotkeys_updated(set(["extra_hotkey"]))

        self.assertEqual(len(self.model_tracker.miner_hotkey_to_model_metadata_dict), 0)
        self.assertEqual(len(self.model_tracker.miner_hotkey_to_eval_results), 0)

    def test_on_model_evaluated_cleared_on_metadata_update(self):
        hotkey = "test_hotkey"
        model_id = self._create_model_id()

        model_metadata = ModelMetadata(id=model_id, block=1)
        self.model_tracker.on_miner_model_updated(hotkey, model_metadata)

        # Record a model evaluation.
        eval_result = EvalResult(
            block=2, score=1, winning_model_block=1, winning_model_score=2
        )
        self.model_tracker.on_model_evaluated(hotkey, 1, eval_result)
        self.assertEqual(
            self.model_tracker.get_eval_results_for_miner_hotkey(hotkey, 1),
            [eval_result],
        )

        # Call on_miner_model_updated with the same model_metadata. This shouldn't clear the eval results.
        self.model_tracker.on_miner_model_updated(hotkey, model_metadata)
        self.assertEqual(
            self.model_tracker.get_eval_results_for_miner_hotkey(hotkey, 1),
            [eval_result],
        )

        # Now update the model metadata with a new metadata. This should clear the eval results.
        self.model_tracker.on_miner_model_updated(
            hotkey, ModelMetadata(id=model_id, block=2)
        )
        self.assertEqual(
            len(self.model_tracker.get_eval_results_for_miner_hotkey(hotkey, 1)), 0
        )

    def test_on_model_evaluated_multiple_miners(self):
        """Verifies that miner evaluations are stored per miner."""
        miner1 = "miner1"
        miner2 = "miner2"

        eval_result1 = EvalResult(
            block=1, score=1, winning_model_block=1, winning_model_score=2
        )
        eval_result2 = EvalResult(
            block=2, score=2, winning_model_block=1, winning_model_score=2
        )
        self.model_tracker.on_model_evaluated(miner1, 1, eval_result1)
        self.model_tracker.on_model_evaluated(miner2, 1, eval_result2)

        self.assertEqual(
            self.model_tracker.get_eval_results_for_miner_hotkey(miner1, 1),
            [eval_result1],
        )
        self.assertEqual(
            self.model_tracker.get_eval_results_for_miner_hotkey(miner2, 1),
            [eval_result2],
        )

    def test_on_model_evaluated_respects_competition_id(self):
        """Verifies that miner evaluations are stored per competition."""

        eval_result1 = EvalResult(
            block=1, score=1, winning_model_block=1, winning_model_score=2
        )
        eval_result2 = EvalResult(
            block=2, score=2, winning_model_block=1, winning_model_score=2
        )
        self.model_tracker.on_model_evaluated("miner", 1, eval_result1)
        self.model_tracker.on_model_evaluated("miner", 1, eval_result2)

        self.assertEqual(
            self.model_tracker.get_eval_results_for_miner_hotkey("miner", 1),
            [eval_result1, eval_result2],
        )
        self.assertEqual(
            self.model_tracker.get_eval_results_for_miner_hotkey("miner", 2), []
        )

    def test_eval_results_are_ordered_and_truncated(self):
        """Verifies that only 3 eval results are kept and that they're returned in block order."""

        miner1 = "miner1"
        eval_results = [
            EvalResult(
                block=i, score=10 - i, winning_model_block=1, winning_model_score=2
            )
            for i in range(50)
        ]
        random.shuffle(eval_results)

        # A bit unusual, but insert the evaluations in a random order.
        # Per the API spec, this should still result in the 3 most recent being kept.
        for eval_result in eval_results:
            self.model_tracker.on_model_evaluated(miner1, 1, eval_result)

        expected = sorted(eval_results, key=lambda er: er.block)[-3:]
        self.assertEqual(
            self.model_tracker.get_eval_results_for_miner_hotkey(miner1, 1), expected
        )

    def test_get_block_last_evaluated(self):
        hotkey = "hotkey"

        # 1. Check we get an empty result when no evals have been performed
        self.assertIsNone(self.model_tracker.get_block_last_evaluated(hotkey))

        # 2. Check we get the correct block when a single competition has been evaluated
        self.model_tracker.on_model_evaluated(
            hotkey,
            1,
            EvalResult(block=2, score=2, winning_model_block=1, winning_model_score=2),
        )
        self.assertEqual(self.model_tracker.get_block_last_evaluated(hotkey), 2)

        # 3. Now add another competition eval at an older block and check we still get the right result.
        self.model_tracker.on_model_evaluated(
            hotkey,
            2,
            EvalResult(block=1, score=1, winning_model_block=1, winning_model_score=2),
        )
        self.assertEqual(self.model_tracker.get_block_last_evaluated(hotkey), 2)

        # 4. Evaluate again on the new competition at a newer block.
        self.model_tracker.on_model_evaluated(
            hotkey,
            2,
            EvalResult(block=3, score=3, winning_model_block=1, winning_model_score=2),
        )
        self.assertEqual(self.model_tracker.get_block_last_evaluated(hotkey), 3)

        # 5. Finally, add another eval for the first competition.
        self.model_tracker.on_model_evaluated(
            hotkey,
            1,
            EvalResult(block=5, score=4, winning_model_block=1, winning_model_score=2),
        )
        self.assertEqual(self.model_tracker.get_block_last_evaluated(hotkey), 5)

    def test_clear_eval_results_missing_comp(self):
        """Verifies that clear_eval_results doesn't crash when the competition doesn't exist."""
        
        eval_result1 = EvalResult(
            block=1, score=1, winning_model_block=1, winning_model_score=2
        )
        eval_result2 = EvalResult(
            block=2, score=2, winning_model_block=1, winning_model_score=2
        )
        self.model_tracker.on_model_evaluated("miner", 1, eval_result1)
        self.model_tracker.on_model_evaluated("miner", 1, eval_result2)

        self.model_tracker.clear_eval_results(2)

        self.assertEqual(
            self.model_tracker.get_eval_results_for_miner_hotkey("miner", 1),
            [eval_result1, eval_result2],
        )

    def test_clear_eval_results(self):
        """Verifies that clear_eval_results only deletes results from the right competition."""

        eval_result1 = EvalResult(
            block=1, score=1, winning_model_block=1, winning_model_score=2
        )
        eval_result2 = EvalResult(
            block=2, score=2, winning_model_block=1, winning_model_score=2
        )
        self.model_tracker.on_model_evaluated("miner", 1, eval_result1)
        self.model_tracker.on_model_evaluated("miner", 1, eval_result1)
        self.model_tracker.on_model_evaluated("miner", 2, eval_result2)

        self.assertEqual(
            self.model_tracker.get_eval_results_for_miner_hotkey("miner", 1),
            [eval_result1, eval_result1],
        )
        self.model_tracker.clear_eval_results(1)
        self.assertEqual(
            self.model_tracker.get_eval_results_for_miner_hotkey("miner", 1), []
        )

if __name__ == "__main__":
    unittest.main()



import os
import tempfile
from dataclasses import replace
from typing import Optional

from huggingface_hub import HfApi
from huggingface_hub.utils import RepositoryNotFoundError
from transformers import AutoModelForCausalLM

from taoverse.model.competition.data import ModelConstraints
from taoverse.model.data import Model, ModelId
from taoverse.model.model_updater import MinerMisconfiguredError
from taoverse.model.storage.disk import utils
from taoverse.model.storage.remote_model_store import RemoteModelStore


class HuggingFaceModelStore(RemoteModelStore):
    """Hugging Face based implementation for storing and retrieving a model."""

    @classmethod
    def assert_access_token_exists(cls) -> str:
        """Asserts that the access token exists."""
        if not os.getenv("HF_ACCESS_TOKEN"):
            raise ValueError("No Hugging Face access token found to write to the hub.")
        return os.getenv("HF_ACCESS_TOKEN")

    @classmethod
    def get_access_token_if_exists(cls) -> Optional[str]:
        """Returns the access token if it exists."""
        return os.getenv("HF_ACCESS_TOKEN")

    async def upload_model(
        self, model: Model, model_constraints: ModelConstraints
    ) -> ModelId:
        """Uploads a trained model to Hugging Face."""
        token = HuggingFaceModelStore.assert_access_token_exists()

        commit_info = model.pt_model.push_to_hub(
            repo_id=model.id.namespace + "/" + model.id.name,
            token=token,
            safe_serialization=True,
            private=True,
        )

        model_id_with_commit = replace(model.id, commit=commit_info.oid)

        # To make sure we get the same hash as validators, we need to redownload it at a
        # local tmp directory after which it can be deleted.
        with tempfile.TemporaryDirectory() as temp_dir:
            model_with_hash = await self.download_model(
                model_id_with_commit, temp_dir, model_constraints
            )
            # Return a ModelId with both the correct commit and hash.
            return model_with_hash.id

    async def download_model(
        self,
        model_id: ModelId,
        local_path: str,
        model_constraints: ModelConstraints,
    ) -> Model:
        """Retrieves a trained model from Hugging Face."""
        if not model_id.commit:
            raise ValueError("No Hugging Face commit id found to read from the hub.")

        repo_id = model_id.namespace + "/" + model_id.name
        token = HuggingFaceModelStore.get_access_token_if_exists()

        # Check ModelInfo for the size of model.safetensors file before downloading.
        api = HfApi()
        try:
            model_info = api.model_info(
                repo_id=repo_id,
                revision=model_id.commit,
                timeout=10,
                files_metadata=True,
                token=token,
            )
        except RepositoryNotFoundError:
            raise MinerMisconfiguredError(
                hotkey="N/A",
                message=f"HuggingFace repository {repo_id} with revision {model_id.commit} was not found on the hub.",
            )

        size = sum(repo_file.size for repo_file in model_info.siblings)
        if size > model_constraints.max_bytes:
            raise MinerMisconfiguredError(
                hotkey="N/A",
                message=f"Hugging Face repo over maximum size limit. Size {size}. Limit {model_constraints.max_bytes}.",
            )

        # Transformers library can pick up a model based on the hugging face path (username/model) + rev.
        try:
            model = AutoModelForCausalLM.from_pretrained(
                pretrained_model_name_or_path=repo_id,
                revision=model_id.commit,
                cache_dir=local_path,
                use_safetensors=True,
                token=token,
                **model_constraints.kwargs,
            )
        except ValueError as e:
            # This is treated as a MinerMisconfiguredError. Since the error in this case,
            # and after the above checks, most probably comes from `kwargs` misconfigured for
            # the type of model being loaded.
            # For example, attempting to use FlashAttention 2.0 with GPT2 models.
            # The latter does not support that. This is an error known for SN9 when trying
            # to load 772M models into the default 7B when migrating to multi-competition support.
            raise MinerMisconfiguredError(
                hotkey="N/A",
                message=f"Model {repo_id}/{model_id.commit} could not be loaded with kwargs {model_constraints.kwargs}, Here is the error trace:",
            ) from e

        # Get the directory the model was stored to.
        model_dir = utils.get_hf_download_path(local_path, model_id)

        # Realize all symlinks in that directory since Transformers library does not support avoiding symlinks.
        utils.realize_symlinks_in_directory(model_dir)

        # Compute the hash of the downloaded model.
        model_hash = utils.get_hash_of_directory(model_dir)
        model_id_with_hash = replace(model_id, hash=model_hash)

        return Model(id=model_id_with_hash, pt_model=model)



import os

import wandb


def login():
    """Logs in to wandb using the access token from the environment."""
    access_token = os.getenv("WANDB_ACCESS_TOKEN")
    if not access_token:
        raise ValueError("No WANDB_ACCESS_TOKEN found in .env")

    wandb.login(key=access_token)



from abc import ABC, abstractmethod


class EpsilonFunc(ABC):
    """
    Interface for a function to compute the current epsilon value.
    """

    @abstractmethod
    def compute_epsilon(self, current_block: int, model_block: int) -> float:
        """
        Computes the epsilon value.

        Args:
            current_block: The current block.
            model_block: The block at which the model was submitted.

        Returns:
            The computed epsilon value.
        """
        pass


class FixedEpsilon(EpsilonFunc):
    """
    A fixed epsilon value.
    """

    def __init__(self, epsilon: float):
        """
        Initializes the FixedEpsilon.

        Args:
            epsilon: The fixed epsilon value.
        """
        self.epsilon = epsilon

    def compute_epsilon(self, current_block: int, model_block: int) -> float:
        return self.epsilon


class LinearDecay(EpsilonFunc):
    """
    An epsilon function linearly decays epsilon to a minimum epsilon over a defined number of blocks.
    """

    def __init__(self, start_epsilon: float, end_epsilon: float, decay_blocks: int):
        """
        Initializes the FixedEpsilon.

        Args:
            epsilon: The fixed epsilon value.
        """
        self.start_epsilon = start_epsilon
        self.end_epsilon = end_epsilon
        self.decay_blocks = decay_blocks

    def compute_epsilon(self, current_block: int, model_block: int) -> float:
        # Find the difference in blocks.
        # In case of metagraph divergance between checking blocks and syncing models default to 0.
        block_difference = max(current_block - model_block, 0)

        # Use a linear progression based on the decay blocks capping at complete decay.
        block_adjustment = min(block_difference / self.decay_blocks, 1)

        epsilon_adjustment = block_adjustment * (self.start_epsilon - self.end_epsilon)

        return self.start_epsilon - epsilon_adjustment



import dataclasses
import functools
import time
import unittest
from tempfile import NamedTemporaryFile

from numpy import str_

from taoverse.utilities import utils
from taoverse.utilities.utils import run_in_subprocess, run_in_thread


@dataclasses.dataclass
class SimpleObject:
    int_field: int
    bool_field: bool
    str_field: str
    list_field: list
    dict_field: dict


class TestUtils(unittest.TestCase):
    def test_run_in_subprocess(self):
        def test_func(a: int, b: int):
            return a + b

        partial = functools.partial(test_func, 1, 2)

        result = run_in_subprocess(func=partial, ttl=5)
        self.assertEqual(3, result)

    def test_run_in_subprocess_timeout(self):
        def test_func(a: int, b: int):
            time.sleep(3)
            return a + b

        partial = functools.partial(test_func, 1, 2)

        with self.assertRaises(TimeoutError):
            result = run_in_subprocess(func=partial, ttl=1)

    def test_run_in_subprocess_no_return(self):
        def test_func(a: int, b: int):
            pass

        partial = functools.partial(test_func, 1, 2)

        result = run_in_subprocess(func=partial, ttl=5)
        self.assertIsNone(result)

    def test_run_in_subprocess_tuple_return(self):
        def test_func(a: int, b: int):
            return a, b

        partial = functools.partial(test_func, 1, 2)

        result = run_in_subprocess(func=partial, ttl=5)
        self.assertEqual((1, 2), result)

    def test_run_in_subprocess_exception(self):
        def test_func(a: int, b: int):
            raise ValueError()

        partial = functools.partial(test_func, 1, 2)

        with self.assertRaises(ValueError):
            result = run_in_subprocess(func=partial, ttl=5)

    def test_run_in_thread(self):
        def test_func(a: int, b: int):
            return a + b

        partial = functools.partial(test_func, 1, 2)

        result = run_in_thread(func=partial, ttl=5)
        self.assertEqual(3, result)

    def test_run_in_thread_timeout(self):
        def test_func(a: int, b: int):
            time.sleep(3)
            return a + b

        partial = functools.partial(test_func, 1, 2)

        with self.assertRaises(TimeoutError):
            result = run_in_thread(func=partial, ttl=1)

    def test_run_in_thread_no_return(self):
        def test_func(a: int, b: int):
            pass

        partial = functools.partial(test_func, 1, 2)

        result = run_in_thread(func=partial, ttl=5)
        self.assertIsNone(result)

    def test_run_in_thread_tuple_return(self):
        def test_func(a: int, b: int):
            return a, b

        partial = functools.partial(test_func, 1, 2)

        result = run_in_thread(func=partial, ttl=5)
        self.assertEqual((1, 2), result)

    def test_run_in_thread_exception(self):
        def test_func(a: int, b: int):
            raise ValueError()

        partial = functools.partial(test_func, 1, 2)

        with self.assertRaises(ValueError):
            result = run_in_thread(func=partial, ttl=5)

    def test_save_and_load_version(self):
        version = 1023  # Typically equivalent to "1.2.3" in our convention.
        with NamedTemporaryFile() as f:
            self.assertIsNone(utils.get_version(f.name))

            utils.save_version(f.name, version)
            self.assertEqual(utils.get_version(f.name), version)

    def test_fingerprint_single_object(self):
        a = SimpleObject(1, True, "hello", [1, 2, 3], {"a": 1, "b": 2})
        self.assertEqual(utils.fingerprint(a), utils.fingerprint(a))

        b = SimpleObject(1, True, "other", [1, 2, 3], {"a": 1, "b": 2})
        self.assertNotEqual(utils.fingerprint(a), utils.fingerprint(b))

    def test_fingerprint_multiple_objects(self):
        a = SimpleObject(1, True, "hello", [1, 2, 3], {"a": 1, "b": 2})
        b = SimpleObject(1, True, "other", [1, 2, 3], {"a": 1, "b": 2})

        self.assertEqual(utils.fingerprint([a, b]), utils.fingerprint([a, b]))
        self.assertEqual(utils.fingerprint([a]), utils.fingerprint([a]))

        self.assertNotEqual(utils.fingerprint([a]), utils.fingerprint([b]))
        self.assertNotEqual(utils.fingerprint([a, b]), utils.fingerprint([b, a]))


if __name__ == "__main__":
    unittest.main()



import unittest

from taoverse.model import utils
from taoverse.model.data import ModelMetadata, ModelId


class TestModelUtils(unittest.TestCase):
    def test_get_hash_of_two_strings(self):
        string1 = "hello"
        string2 = "world"

        result = utils.get_hash_of_two_strings(string1, string2)

        self.assertEqual(result, "k2oYXKqiZrucvpgengXLeM1zKwsygOuURBK7b4+PB68=")

    def test_validate_hf_repo_id_too_long(self):
        with self.assertRaises(ValueError) as ve:
            # Max allowed length is 41 characters
            utils.validate_hf_repo_id("my-org/" + "a" * 40)

        self.assertRegex(
            str(ve.exception),
            "Hugging Face repo id must be between 3 and 38 characters",
        )

    def test_validate_hf_repo_id_incorrect_format(self):
        with self.assertRaises(ValueError) as ve:
            utils.validate_hf_repo_id("my-repo-name-without-a-namespace")

        self.assertRegex(
            str(ve.exception), "must be in the format <org or user name>/<repo_name>"
        )

    def test_validate_hf_repo_id_valid(self):
        namespace, name = utils.validate_hf_repo_id("my-org/my-repo-name")
        self.assertEqual("my-org", namespace)
        self.assertEqual("my-repo-name", name)

    def test_get_hf_repo(self):
        metadata = ModelMetadata(
            id=ModelId(
                namespace="my-org",
                name="my-repo-name",
                commit="test-commit",
                hash="test-hash",
                secure_hash="test-secure-hash",
                competition_id=1,
            ),
            block=0,
        )
        repo_name = utils.get_hf_repo_name(metadata)
        self.assertEqual("my-org/my-repo-name", repo_name)

    def test_get_hf_url(self):
        metadata = ModelMetadata(
            id=ModelId(
                namespace="my-org",
                name="my-repo-name",
                commit="test-commit",
                hash="test-hash",
                secure_hash="test-secure-hash",
                competition_id=1,
            ),
            block=0,
        )
        url = utils.get_hf_url(metadata)
        self.assertEqual(
            "https://huggingface.co/my-org/my-repo-name/tree/test-commit", url
        )


if __name__ == "__main__":
    unittest.main()



import asyncio
import dataclasses
import functools
import threading
import traceback
from dataclasses import field
from datetime import datetime
from typing import Callable, Dict, List, Optional

import bittensor as bt

from taoverse.utilities import utils


class MetagraphSyncer:
    @dataclasses.dataclass
    class _State:
        metagraph: Optional[bt.metagraph] = None
        last_synced_time: Optional[datetime] = None
        listeners: List = field(default_factory=list)

    def __init__(
        self, subtensor: bt.subtensor, config: Dict[int, int], lite: bool = True
    ):
        """Constructs a new MetagraphSyncer, that periodically refreshes metagraph defined in the config.

        Args:
            subtensor (bt.subtensor): The subtensor used to fetch the metagraphs.
            config (Dict[int, int]): A mapping of netuid to the cadence (in seconds) to sync the metagraph.
            lite (bool): If the metagraphs should be synced as a lite metagraph (without detailed Weights).
        """
        self.subtensor = subtensor
        self.config = config
        self.lite = lite
        self.metagraph_map: Dict[int, MetagraphSyncer._State] = {
            netuid: MetagraphSyncer._State() for netuid in config.keys()
        }
        self.is_running = False
        self.done_initial_sync = False
        self.lock = threading.RLock()

        bt.logging.info(f"MetagraphSyncer created with config: {config}")

    def do_initial_sync(self):
        """Performs an initial sync of all metagraphs.

        Unlike regular syncs, this will not notify listeners of the updated metagraph.
        """
        bt.logging.debug("Metagraph syncer do_initial_sync called")

        for netuid in self.config.keys():
            fn = functools.partial(self.subtensor.metagraph, netuid, self.lite)
            metagraph = utils.run_in_thread(fn, ttl=120, name=f"InitalSync-{netuid}")
            with self.lock:
                state = self.metagraph_map[netuid]
                state.metagraph = metagraph
                state.last_synced_time = datetime.now()

            bt.logging.debug(
                f"Successfully loaded metagraph for {netuid} with lite: {self.lite}."
            )

        self.done_initial_sync = True

    def start(self):
        bt.logging.debug("Metagraph syncer start called")

        assert self.done_initial_sync, "Must call do_initial_sync before starting"

        self.is_running = True
        thread = threading.Thread(target=self._run, daemon=True)
        thread.start()

    async def _sync_metagraph_loop(self, netuid: int, cadence: int):
        while self.is_running:
            # On start, wait cadence before the first sync.
            bt.logging.trace(f"Syncing metagraph for {netuid} in {cadence} seconds.")
            await asyncio.sleep(cadence)

            try:
                # Intentionally block the shared thread so that we only
                # sync 1 metagraph at a time.
                bt.logging.trace(f"Syncing metagraph for {netuid}.")
                metagraph = utils.run_in_thread(
                    functools.partial(self.subtensor.metagraph, netuid, self.lite),
                    ttl=120,
                    name=f"Sync-{netuid}",
                )
                bt.logging.trace(
                    f"Successfully synced metagraph for {netuid} with lite: {self.lite}."
                )
                state = None
                with self.lock:
                    # Store metagraph and sync time
                    state = self.metagraph_map[netuid]
                    state.metagraph = metagraph
                    state.last_synced_time = datetime.now()

                self._notify_listeners(state, netuid)
            except (BaseException, Exception) as e:
                bt.logging.error(
                    f"Error when syncing metagraph for {netuid}: {e}. Retrying in 60 seconds."
                )
                await asyncio.sleep(60)

    async def _run_async(self):
        # For each netuid we should sync metagraphs for, spawn a Task to sync it.
        await asyncio.wait(
            [
                asyncio.create_task(self._sync_metagraph_loop(netuid, cadence))
                for netuid, cadence in self.config.items()
            ],
            return_when=asyncio.ALL_COMPLETED,
        )

    def _run(self):
        try:
            asyncio.run(self._run_async())
        finally:
            bt.logging.info("MetagraphSyncer _run complete.")

    def register_listener(
        self, listener: Callable[[bt.metagraph, int], None], netuids: List[int]
    ):
        """Registers a listener to be notified when a metagraph for any netuid in netuids is updated.

        The listener will be called from a different thread, so it must be thread-safe.
        """
        if not netuids:
            raise ValueError("Must provide at least 1 netuid")

        with self.lock:
            for netuid in netuids:
                if netuid not in self.metagraph_map:
                    raise ValueError(
                        f"Metagraph for {netuid} not being tracked in MetagraphSyncer."
                    )
                self.metagraph_map[netuid].listeners.append(listener)

    def get_metagraph(self, netuid: int) -> bt.metagraph:
        """Returns the last synced version of the metagraph for netuid."""
        with self.lock:
            if netuid not in self.metagraph_map:
                raise ValueError(
                    f"Metagraph for {netuid} not known to MetagraphSyncer."
                )
            metagraph = self.metagraph_map[netuid].metagraph
            if not metagraph:
                raise ValueError(f"Metagraph for {netuid} has not been synced yet.")
            return metagraph

    def _notify_listeners(self, state: _State, netuid: int):
        """Notifies listeners of a new metagraph for netuid."""
        bt.logging.debug(f"Notifying listeners of update to metagraph for {netuid}.")

        for listener in state.listeners:
            try:
                listener(state.metagraph, netuid)
            except Exception:
                bt.logging.error(
                    f"Exception caught notifying {netuid} listener of metagraph update.\n{traceback.format_exc()}"
                )



from typing import List, Optional, Tuple

from taoverse.model.competition.data import Competition


def get_competition_for_block(
    comp_id: int,
    block: int,
    schedule_by_block: List[Tuple[int, List[Competition]]],
) -> Optional[Competition]:
    """Returns the competition for the given id at the given block, or None if it does not exist."""
    competition_schedule = get_competition_schedule_for_block(block, schedule_by_block)
    for comp in competition_schedule:
        if comp.id == comp_id:
            return comp
    return None


def get_competition_schedule_for_block(
    block: int, schedule_by_block: List[Tuple[int, List[Competition]]]
) -> List[Competition]:
    """Returns the competition schedule at block."""
    competition_schedule = None
    for b, schedule in schedule_by_block:
        if block >= b:
            competition_schedule = schedule
    assert (
        competition_schedule is not None
    ), f"No competition schedule found for block {block}"
    return competition_schedule



import abc

from taoverse.model.competition.data import ModelConstraints
from taoverse.model.data import Model, ModelId


class RemoteModelStore(abc.ABC):
    """An abstract base class for storing and retrieving a pre trained model."""

    @abc.abstractmethod
    async def upload_model(
        self, model: Model, model_constraints: ModelConstraints
    ) -> ModelId:
        """Uploads a trained model in the appropriate location based on implementation."""
        pass

    @abc.abstractmethod
    async def download_model(
        self,
        model_id: ModelId,
        local_path: str,
        model_constraints: ModelConstraints,
    ) -> Model:
        """Retrieves a trained model from the appropriate location and stores at the given path."""
        pass



from enum import IntEnum
import math


class NormalizationId(IntEnum):
    """Enumeration of normalization methods."""

    # No normalization is applied and the raw score is used.
    NONE = 0

    # Normalizes between [0, 1] using an inverse exponential function.
    INVERSE_EXPONENTIAL = 1


def normalize_score(
    score: float,
    normalization_id: NormalizationId,
    norm_kwargs: dict = {},
) -> float:
    """Normalizes a score based on the provided normalization method.

    Args:
        score (float): The raw score to normalize.
        normalization_id (NormalizationId): The identifier of the normalization method to use.
        norm_kwargs (dict): Keyword arguments to pass to the normalization method.

    Returns:
        float: The normalized score.
    """
    match normalization_id:
        case NormalizationId.NONE:
            return _normalize_none(score)
        case NormalizationId.INVERSE_EXPONENTIAL:
            return _normalize_inverse_exponential(score, **norm_kwargs)
        case _:
            raise ValueError(f"Unhandled normalization method {normalization_id}.")


def _normalize_inverse_exponential(score: float, ceiling: float) -> float:
    """Normalizes between [0, 1] using an inverse exponential function.

    Args:
        score (float): The raw score to normalize.
        ceiling (float): The maximum desirable score. Everything above this will be clipped.

    Returns:
        float: The normalized score.
    """
    if score >= ceiling:
        return 1.0

    return (1 - math.exp(-score / ceiling)) / (1 - math.exp(-1))


def _normalize_none(score: float) -> float:
    return score



import random
import unittest

from taoverse.model.eval.normalization import NormalizationId, normalize_score


class TestNormalizeScore(unittest.TestCase):

    def test_normalize_none(self):
        for _ in range(10):
            score = random.random()
            normalization_id = NormalizationId.NONE
            normalized_score = normalize_score(score, normalization_id)
            self.assertEqual(normalized_score, score)

    def test_normalize_inverse_exponential(self):
        score = 5.0
        normalization_id = NormalizationId.INVERSE_EXPONENTIAL
        norm_kwargs = {"ceiling": 10.0}
        normalized_score = normalize_score(score, normalization_id, norm_kwargs)
        self.assertAlmostEqual(normalized_score, 0.6224593312018546, places=6)

    def test_normalize_inverse_exponential_above_ceiling(self):
        score = 15.0
        normalization_id = NormalizationId.INVERSE_EXPONENTIAL
        norm_kwargs = {"ceiling": 10.0}
        normalized_score = normalize_score(score, normalization_id, norm_kwargs)
        self.assertEqual(normalized_score, 1.0)

    def test_normalize_inverse_exponential_always_between_0_and_1(self):
        for i in range(101):
            score = i / 10
            normalization_id = NormalizationId.INVERSE_EXPONENTIAL
            norm_kwargs = {"ceiling": 10.0}
            normalized_score = normalize_score(score, normalization_id, norm_kwargs)
            self.assertTrue(0 <= normalized_score <= 1)

    def test_unhandled_normalization_method(self):
        score = 5.0
        normalization_id = 999  # Invalid normalization method
        norm_kwargs = {}
        with self.assertRaises(ValueError):
            normalize_score(score, normalization_id, norm_kwargs)


if __name__ == "__main__":
    unittest.main()



import threading
import unittest
from unittest import mock

import bittensor as bt

from taoverse.metagraph.metagraph_syncer import MetagraphSyncer


class TestMetagraphSyncer(unittest.TestCase):
    def test_do_initial_sync(self):
        # Mock subtensor.metagraph() function
        metagraph1 = bt.metagraph(netuid=1, sync=False)
        metagraph2 = bt.metagraph(netuid=2, sync=False)

        def get_metagraph(netuid, lite) -> bt.metagraph:
            if netuid == 1:
                return metagraph1
            elif netuid == 2:
                return metagraph2
            else:
                raise Exception("Invalid netuid")

        mock_subtensor = mock.MagicMock(spec=bt.subtensor)
        metagraph_mock = mock.MagicMock(side_effect=get_metagraph)
        mock_subtensor.metagraph = metagraph_mock

        # Create MetagraphSyncer instance with mock subtensor
        metagraph_syncer = MetagraphSyncer(mock_subtensor, {1: 1, 2: 1})

        # Call do_initial_sync method
        metagraph_syncer.do_initial_sync()

        # Verify get_metagraph() returns the expected metagraph.
        # We can't check object equality because of how equality is done on bt.metagraph
        # so just check the netuid.
        self.assertEqual(metagraph_syncer.get_metagraph(1).netuid, metagraph1.netuid)
        self.assertEqual(metagraph_syncer.get_metagraph(2).netuid, metagraph2.netuid)

    def test_listener_called(self):
        # Mock subtensor.metagraph() function
        metagraph1 = bt.metagraph(netuid=1, sync=False)
        metagraph2 = bt.metagraph(netuid=2, sync=False)

        def get_metagraph(netuid, lite) -> bt.metagraph:
            if netuid == 1:
                return metagraph1
            elif netuid == 2:
                return metagraph2
            else:
                raise Exception("Invalid netuid")

        mock_subtensor = mock.MagicMock(spec=bt.subtensor)
        metagraph_mock = mock.MagicMock(side_effect=get_metagraph)
        mock_subtensor.metagraph = metagraph_mock

        # Create MetagraphSyncer instance with mock subtensor
        metagraph_syncer = MetagraphSyncer(mock_subtensor, {1: 1, 2: 1})

        # Call do_initial_sync method
        metagraph_syncer.do_initial_sync()

        # Register a listener for netuid 1.
        event = threading.Event()

        def listener(metagraph, netuid):
            self.assertEqual(metagraph.netuid, 1)
            self.assertEqual(netuid, 1)
            event.set()

        metagraph_syncer.register_listener(listener, [1])

        # Since we sync every 1 second, verify the listener is called within 5 seconds.
        event.wait(5)

    def test_lite_metagraph(self):
        # Mock subtensor.metagraph() function
        metagraph1 = bt.metagraph(netuid=1, sync=False)

        def get_metagraph(netuid, lite) -> bt.metagraph:
            if netuid == 1 and lite == True:
                return metagraph1
            else:
                raise Exception("Invalid netuid")

        mock_subtensor = mock.MagicMock(spec=bt.subtensor)
        metagraph_mock = mock.MagicMock(side_effect=get_metagraph)
        mock_subtensor.metagraph = metagraph_mock

        # Create MetagraphSyncer instance with mock subtensor
        metagraph_syncer = MetagraphSyncer(
            mock_subtensor,
            {
                1: 1,
            },
        )

        # Call do_initial_sync method
        metagraph_syncer.do_initial_sync()

        # Verify get_metagraph() returns the expected metagraph.
        self.assertEqual(metagraph_syncer.get_metagraph(1).netuid, metagraph1.netuid)

    def test_non_lite_metagraph(self):
        # Mock subtensor.metagraph() function
        metagraph1 = bt.metagraph(netuid=1, sync=False)

        def get_metagraph(netuid, lite) -> bt.metagraph:
            if netuid == 1 and lite == False:
                return metagraph1
            else:
                raise Exception("Invalid netuid")

        mock_subtensor = mock.MagicMock(spec=bt.subtensor)
        metagraph_mock = mock.MagicMock(side_effect=get_metagraph)
        mock_subtensor.metagraph = metagraph_mock

        # Create MetagraphSyncer instance with mock subtensor
        metagraph_syncer = MetagraphSyncer(mock_subtensor, {1: 1}, lite=False)

        # Call do_initial_sync method
        metagraph_syncer.do_initial_sync()

        # Verify get_metagraph() returns the expected metagraph.
        self.assertEqual(metagraph_syncer.get_metagraph(1).netuid, metagraph1.netuid)


if __name__ == "__main__":
    unittest.main()



import unittest

from taoverse.model.eval.normalization import NormalizationId
from taoverse.model.eval.task import EvalTask


class TestEvalTask(unittest.TestCase):

    def test_eval_task_initialization(self):
        task = EvalTask(
            name="Test Task",
            dataset_id=1,
            method_id=1,
            normalization_id=NormalizationId.NONE,
        )
        self.assertEqual(task.name, "Test Task")
        self.assertEqual(task.dataset_id, 1)
        self.assertEqual(task.method_id, 1)
        self.assertEqual(task.normalization_id, NormalizationId.NONE)
        self.assertEqual(task.weight, 1.0)

    def test_eval_task_weight_validation(self):
        with self.assertRaises(ValueError):
            EvalTask(
                name="Test Task",
                dataset_id=1,
                method_id=2,
                normalization_id=NormalizationId.NONE,
                weight=0,
            )

    def test_eval_task_normalization_kwargs_validation_none(self):
        with self.assertRaises(ValueError):
            EvalTask(
                name="Test Task",
                dataset_id=1,
                method_id=1,
                normalization_id=NormalizationId.NONE,
                normalization_kwargs={"some_key": "some_value"},
            )

    def test_eval_task_normalization_kwargs_validation_inverse_exponential(self):
        with self.assertRaises(ValueError):
            EvalTask(
                name="Test Task",
                dataset_id=1,
                method_id=2,
                normalization_id=NormalizationId.INVERSE_EXPONENTIAL,
            )


if __name__ == "__main__":
    unittest.main()
